'use strict'

const util = require('util')
const Core = require('../core')
const Schemas = require('../Schemas')

/**
*
* Base Transport
*
*/
class BaseTransport extends Core {
  constructor (options) {
    super(options)

    this._subscription = {}

    if (this.options.schemas) {
      this.setSchemas(this.options.schemas)
    }

    // The clients can be used both ways, but we need to
    // know what their role is, normal role for a client
    // is being the runtime, runtime will be refered to as `server` role
    // a UI or whatever commander will be refered to as `client` role
    // default role is `server`
    if (this.options.role === 'client') {
      this.receiveMode = 'output'
      this.replyMode = 'input'
      if (!this.options.secret) {
        throw Error('In client mode options requires a secret')
      }
    } else {
      this.receiveMode = 'input'
      this.replyMode = 'output'
    }

    this.connections = {}
  }

  sendAll (protocol, command, data, graphId) {
    const self = this

    // data would contain `graph`, topic tells us we need graph.
    // TODO: merge this with this.connections.
    graphId = graphId || data.graph
    const topic = 'graph'
    if (this._subscription.hasOwnProperty(topic)) {
      if (this._subscription[topic].hasOwnProperty(graphId)) {
        this._subscription[topic][graphId].forEach(function (conn) {
          if (command === 'error') {
            self.sendError(protocol, data, conn)
          } else {
            self.$send({
              protocol: protocol,
              command: command,
              payload: data
            }, conn)
          }
        })
      } else {
        // For now just do not send back.
        // subgraphs are not subscribed, so just do not send info about them
        // e.g. add link will also trigger for subcomponents beneath.
        // The question is, do we need to know about those additions at all.
        // I think not..
        console.log('protocol:' + protocol, 'topic:' + topic, data)
        this.logger.warn('BaseTransport',
          util.format('`%s` property `%s` not found', topic, graphId)
        )
      }
    }
  }

  /**
   *
   * Internal Receive
   *
   * Emit's the validated input for processing.
   *
   */
  receive (data, conn) {
    this.logger.debug('BaseTransport', 'Receive:', data)

    const ret = this.validate(
      this.receiveMode,
      data.protocol,
      data.command,
      data.payload
    )

    if (util.isError(ret)) {
      const msg = {
        message: ret.toString(),
        graph: data.payload.graph
      }
      this.logger.error('BaseTransport', 'Error:', msg)
      this.sendError(data.protocol || 'runtime', msg, conn)
    // this.$send(msg, conn)
    } else {
      // this should be generic
      // Graph subscription
      if (data.protocol === 'graph' && data.command === 'clear') {
        this.subscribe(conn, 'graph', data.payload.id)
      } else if (data.protocol === 'network' && data.command === 'getstatus') {
        this.subscribe(conn, 'graph', data.payload.graph)
      }

      this.emit(data.protocol, data, conn)
    }
  }

  /**
   *
   * Validates incomming and outgoing messages.
   *
   */
  validate (type, protocol, command, data) {
    if (!this.protocolSchema) {
      throw Error('Schemas required')
    }
    return this.protocolSchema.validate(type, protocol, command, data)
  }

  setSchemas (schemas) {
    this.protocolSchema = new Schemas()
    this.protocolSchema.load(schemas)
  }

  /**
   *
   * Generic method to send out the errors
   * Can be used during extention to intercept all errors.
   *
   * @param protocol
   * @param payload
   * @param conn
   */
  sendError (protocol, payload, conn) {
    const msg = {
      protocol: protocol,
      command: 'error',
      payload: payload
    }
    this.$send(msg, conn)
  }

  /**
   *
   * Internal reply method.
   *
   * Does the generic stuff then uses the send method of the transport.
   *
   */
  reply (protocol, command, data, conn) {
    if (data === undefined) {
      throw Error('data must not be undefined')
    }

    /*
     if (data.command === 'error' && util.isError(data.payload.message)) {
     data.payload.message.toString()
     }
     */
    if (data.command === 'error') {
      this.sendError(data.protocol, data.payload, conn)
    } else {
      // too many messages
      // solve this with something like debug: true
      if (command !== 'component') {
        this.logger.debug(
          'BaseTransport',
          'Respond',
          protocol + ':' + command,
          data
        )
      }

      const ret = this.validate(this.replyMode, protocol, command, data)
      if (util.isError(ret)) {
        if (this.options.bail) {
          throw ret
        } else {
          this.logger.warn('BaseTransport', ret)
          this.sendError(protocol, ret, conn)
        // this.$send(msg, conn)
        }
      } else {
        const msg = {
          protocol: protocol,
          command: command,
          payload: data
        }
        this.$send(msg, conn)
      }
    }
  }

  $send (/* data, conn */) {
    throw Error('Send method not implemented')
  }

  getAll (protocol) {
    return this.connections[protocol || 'default'] || []
  }

  register (conn, protocol) {
    protocol = protocol || 'default'
    if (!this.connections.hasOwnProperty(protocol)) {
      this.connections[protocol] = []
    }
    this.connections[protocol].push(conn)
    this.emit('register', conn)
  }

  isRegistered (conn, protocol) {
    protocol = protocol || 'default'
    return this.connections[protocol] &&
    this.connections[protocol].indexOf(conn) >= 0
  }

  unregister (conn, protocol) {
    protocol = protocol || 'default'
    if (this.connections[protocol].indexOf(conn) === -1) {
      return
    }

    this.connections[protocol].splice(
      this.connections[protocol].indexOf(conn),
      1
    )
    this.emit('unregister', conn)
  // unsubscribe from everything
  }

  /**
   *
   * Used to subscribe a connection to a key value pair.
   *
   * 'graph': [<id>, <id> ]
   *
   * transport.subscribe(conn, 'graph', value)
   *
   * This method is called by handlers so when the monitor does a sendAll
   * the transport is able to determine who is all.
   *
   * Subscribe is done during graph creation.
   *
   */
  subscribe (conn, key, value) {
    if (!this._subscription.hasOwnProperty(key)) {
      this._subscription[key] = {}
    }

    if (!this._subscription[key].hasOwnProperty(value)) {
      this._subscription[key][value] = []
    }

    if (this._subscription[key][value].indexOf(conn) === -1) {
      this._subscription[key][value].push(conn)
    }

    this.emit('subscribe', conn, key, value)

    this.logger.info(
      'BaseTransport',
      'Connection subscribed to %s %s',
      key,
      value
    )
  }

  isSubscribed (conn, key, value) {
    if (!this._subscription.hasOwnProperty(key)) {
      return false
    }
    if (!this._subscription[key].hasOwnProperty(value)) {
      return false
    }
    return this._subscription[key][value].indexOf(conn) >= 0
  }

  unsubscribe (conn, key, value) {
    if (this._subscription.hasOwnProperty(key) &&
      this._subscription[key].hasOwnProperty(value)) {
      this._subscription[key][value].splice(
        this._subscription[key][value].indexOf(conn),
        1
      )

      this.emit('unsubscribe', conn, key, value)

      this.logger.info(
        'BaseTransport',
        'Connection unsubscribed from %s %s',
        key,
        value
      )
    }
  }

  setSecret (data) {
    if (this.options.secret) {
      if (!data.payload) {
        data.payload = {}
      }
      data.payload.secret = this.options.secret
    }
  }
}

module.exports = BaseTransport
