'use strict'

const BaseTransport = require('./base')
const httpServer = require('http')
const WebSocketServer = require('websocket').server

/**
*
* Websocket Transport
*
* Usage:
*
*  const nm = new NetworkMonitor({
*    transport: new WebSocketTransport({ logger: console }),
*    logger: console
*  })
*  nm.monitor(ioHandler)
*
* @params {Server} http HTTP Server
* @params {Object} options Options
* @params {String} options.catchExceptions Send all exceptions to UI.
* @params {String} options.type runtime type chix
*/
class WebSocketTransport extends BaseTransport {
  // constructor(http, options) {
  constructor (options, http) {
    // TODO: http as first param is a bit weird
    super(options)
    const self = this

    if (!http) {
      http = httpServer.createServer((req, res) => {
        res.writeHead(200, {
          'Content-Type': 'application/json'
        })
        res.end(JSON.stringify({
          connections: this.connections.length,
          pendingRequests: this.pendingRequests.length
        }))
      })
      self.logger.info('wsServer', 'created http server')
    }

    this.options = options

    this.ws = new WebSocketServer({
      httpServer: http
    })

    this.ws.on('request', function (req) {
      let conn
      try {
        conn = req.accept('noflo', req.origin)
        self.logger.info('wsServer', 'connection accepted', conn.protocol)
        console.log(conn.protocol)
      } catch (e) {
        console.log(e)
        self.logger.error('wsServer', e)
        return
      }

      self.register(conn)

      conn.on('message', function (message) {
        if (message.type === 'utf8') {
          const payload = JSON.parse(message.utf8Data)
          self.receive(payload, conn)
        } else {
          self.logger.warn('wsServer', 'unknown message type', message)
        }
      })

      conn.on('close', function (message) {
        self.logger.info('wsServer', 'connection close', message)

        self.unregister(conn)
      })
    })

    if (options.catchExceptions) {
      process.on('uncaughtException', function (err) {
        // todo: send all will not work here,
        // subscription is always about a graph
        this.sendAll('network', 'error', {
          message: err.toString()
        })
      })
    }
  }

  noCircularJSON (data) {
    return JSON.stringify(data, function (key, val) {
      if (val &&
        typeof val === 'object' &&
        !Array.isArray(val) &&
        val.constructor &&
        val.constructor.name !== 'Object') {
        return '[' + val.constructor.name + ']'
      }
      return val
    })
  }

  /**
   *
   * Send the response.
   *
   */
  $send (data, conn) {
    if (data.command === 'error') {
      if (data.payload.message && data.payload.message.message) {
        data.payload.message = data.payload.message.message
      }
    }
    conn.sendUTF(this.noCircularJSON(data))
  }
}

module.exports = WebSocketTransport
