'use strict'

const WebSocketClient = require('websocket/lib/WebSocketClient')
const BaseTransport = require('./base')

/**
*
* Websocket Client Transport to be used on the server
*
* Usage:
*
*  const nm = new NetworkMonitor({
*    transport: new WebSocketClientTransport({ logger: console }),
*    logger: console
*  })
*  nm.monitor(ioHandler)
*
*  Note: sendAll just keeps on working although we are always
*        just one connection, we could register to several server though.
*
* @params {Object} options Options
* @params {String} options.host ws hostname
* @params {String} options.port ws port
* @params {String} options.mount ws `mount` point
* @params {String} options.protocol noflo
*/
class WebSocketClientTransport extends BaseTransport {
  constructor (options) {
    super(options)

    const self = this

    if (!options.host) {
      throw Error('Must provide a hostname')
    }

    if (!options.port) {
      throw Error('Must provide a portname')
    }

    options.protocol = options.protocol || 'noflo'

    this.options = options

    const url = [
      options.secure ? 'wss://' : 'ws://', options.host, ':', options.port, '/', options.mount || ''
    ].join('')

    this.ws = new WebSocketClient()
    this.ws.on('connect', function (conn) {
      self.logger.info('wsClient', 'WebSocket client connected', conn.protocol)

      // self.register(conn)

      conn.on('message', function (message) {
        if (message.type === 'utf8') {
          const payload = JSON.parse(message.utf8Data)
          self.receive(payload, conn)
        } else {
          self.logger.warn('wsClient', 'unknown message type', message)
        }
      })

      conn.on('error', function (error) {
        self.logger.error('Connection Error: ' + error.toString())
      })

      conn.on('close', function (message) {
        self.logger.info('wsClient', 'connection close', message)

      // self.unregister(conn)
      })
    })

    this.ws.on('connectFailed', function (error) {
      self.logger.warn('wsClient', 'Connect Error: ' + error.toString())
    })

    this.ws.connect(url, options.protocol)
  }

  /**
   *
   * Send the response.
   *
   */
  $send (data, conn) {
    this.setSecret(data)
    conn.sendUTF(JSON.stringify(data))
  }
}

module.exports = WebSocketClientTransport
