'use strict'

const websocket = require('websocket')
const util = require('util')
const BaseTransport = require('./base')
const printmsg = require('../util').printmsg

const WebSocketServer = websocket.server
const WebSocketRouter = websocket.router

/**
*
* Websocket Transport Proxy
*
* Can be used standalone.
*
* The proxied result is available at /proxy
*
* UI talks to /
*
* Client who want's to proxy talks to /proxy
*
* Usage:
*
*  new WebSocketProxyTransport({ logger: console }),
*
* @params {Server} http HTTP Server
* @params {Object} options Options
* @params {String} options.catchExceptions Send all exceptions to UI.
* @params {String} options.type runtime type chix
*/
class WebSocketProxyTransport extends BaseTransport {
  constructor (http, options) {
    super(options)

    const self = this

    this.options = options

    this.ws = new WebSocketServer({
      httpServer: http
    })

    this.router = new WebSocketRouter()
    this.router.attachServer(this.ws)

    // The UI talking
    this.router.mount('*', 'noflo', function (req) {
      const conn = req.accept(req.origin)

      self.logger.info(
        'wsProxyServer',
        'connection accepted from',
        req.origin
      )

      self.register(conn, conn.protocol)

      // send to the other mountpoint subscribers
      conn.on('message', function (message) {
        if (message.type === 'utf8') {
          const data = JSON.parse(message.utf8Data)

          self.logger.debug('wsProxyServer',
            'receveived %s:%s', data.protocol, data.command
          )

          const all = self.getAll('noflo-runtime')

          if (all.length === 0) {
            self.logger.debug(
              'wsProxyServer',
              'request ignored, client list empty'
            )
          } else {
            // Send this info to all graphRunners
            all.forEach(function (client) {
              // there is info in the schema.
              printmsg(self.logger, data, '  → ', client.remoteAddress)
              self.send(data, client)
            })
          }

        // we don't receive.
        // self.receive(payload, conn)
        } else {
          self.logger.warn('wsProxyServer',
            'unknown message type', conn.protocol, message
          )
        }
      })

      conn.on('close', function (message) {
        self.logger.info('wsProxyServer',
          util.format('connection from %s closed', conn.remoteAddress),
          message
        )

        self.unregister(conn, conn.protocol)
      })
    })

    // The Runtime within the browser talking
    this.router.mount('*', 'noflo-runtime', function (req) {
      const conn = req.accept(req.origin)

      // Graph Runners.
      self.register(conn, conn.protocol)

      self.logger.info(
        'wsProxyServer',
        'connection accepted from',
        conn.remoteAddress
      )

      // send to the other mountpoint subscribers
      // This will work, but with multiple it will not ofcourse.
      // Anyway let's just send every message from every client back.
      conn.on('message', function (message) {
        if (message.type === 'utf8') {
          const data = JSON.parse(message.utf8Data)

          // Send this info to all graphRunners
          self.getAll('noflo').forEach(function (client) {
            if (data.command !== 'component') {
              printmsg(self.logger, data, '←   ', client.remoteAddress)
            }
            self.send(data, client)
          })

        // we don't receive.
        // self.receive(payload, conn)
        } else {
          self.logger.warn(
            'wsProxyServer',
            'unknown message type',
            conn.protocol,
            message
          )
        }
      })

      conn.on('close', function (message) {
        self.logger.info(
          'wsProxyServer',
          util.format('connection from %s closed', conn.remoteAddress),
          message
        )

        self.unregister(conn, conn.protocol)
      })
    })
  }

  /**
   *
   * Send the response.
   *
   */
  $send (data, conn) {
    conn.sendUTF(JSON.stringify(data))
  }

  static create (http, options) {
    return new WebSocketProxyTransport(http, options)
  }
}

module.exports = WebSocketProxyTransport
