'use strict'

const BaseTransport = require('./base')
const WebSocketClient = require('websocket/lib/WebSocketClient')
const httpServer = require('http')
const printmsg = require('../util').printmsg
const WebSocketServer = require('websocket').server

function defaultConductor (transport) {
  return {
    send: (payload, clientConn/* , runtimeConn */) => transport.send(payload, clientConn),
    recv: (payload, runtimeConn/* , clientConn */) => transport.send(payload, runtimeConn)
  }
}

/**
*
* Websocket Proxy Server Transport
*
* Will try to connect to a runtime server.
* If it succeeds will start a websocket server which will listen for incomming messages.
*
* Each message received will directly be send to the attached runtime.
*
* Any messages received from the runtime will be send back directly to our client (The UI)
*
* @params {Server} http HTTP Server
* @params {Object} options Options
* @params {String} options.url Websocket url runtime.
* @params {String} options.protocol Protocol noflo.
* @params {String} options.catchExceptions Send all exceptions to UI.
* @params {String} options.type runtime type chix
* @params {String} options.conductor Message conductor Send & Receive negotiator
* @params {HttpServer} http Http Server instance
*/
class WebSocketProxyServerTransport extends BaseTransport {
  constructor (options, http) {
    super(options)
    const self = this

    if (!options.url) {
      throw Error('WebSocketProxyServerTransport: runtime url option required')
    }

    this.conductor = options.conductor
      ? options.conductor(this)
      : defaultConductor(this)

    if (!http) {
      http = httpServer.createServer((req, res) => {
        res.writeHead(200, {
          'Content-Type': 'application/json'
        })
        res.end(JSON.stringify({
          connections: this.connections.length,
          pendingRequests: this.pendingRequests.length
        }))
      })
      self.logger.info('wsServer', 'created http server')
    }

    options.protocol = options.protocol || 'noflo'

    this.options = options

    // start the client first, the one connecting to noflo-nodejs

    // cannot use this transport, because it will try to .receive()
    // so just use bare
    this.client = new WebSocketClient()

    this.conn = null
    this.serverConn = null

    startServer()

    // there is always just one client
    this.client.on('connect', function (conn) {
      self.conn = conn

      self.logger.info('wsClient', 'WebSocket client connected', self.conn.remoteAddress)

      self.conn.on('message', function (message) {
        if (message.type === 'utf8') {
          const payload = JSON.parse(message.utf8Data)
          printmsg(self.logger, payload, '←   ', conn.remoteAddress)
          self.emit('send', payload, conn.remoteAddress)

          self.conductor.send(payload, self.serverConn, conn)
          // self.send(payload, self.serverConn)
        } else {
          self.logger.warn('wsClient', 'unknown message type', message)
        }
      })

      self.conn.on('error', function (error) {
        self.logger.error('Connection Error: ' + error.toString())
      })

      self.conn.on('close', function (message) {
        self.logger.info('wsClient', 'connection close', message)
        // self.unregister(conn)
      })
    })

    this.client.on('connectFailed', function (error) {
      self.logger.warn('wsClient', 'Connect Error: ' + error.toString())
    })

    this.client.connect(options.url, options.protocol)
    // need the send back part

    function startServer () {
      self.server = new WebSocketServer({
        httpServer: http
      })

      self.server.on('request', function (req) {
        let conn
        try {
          conn = req.accept('noflo', req.origin)
          self.serverConn = conn
          self.logger.info('wsServer', 'connection accepted', conn.remoteAddress)
        } catch (e) {
          console.log(e)
          self.logger.error('wsServer', e)
          return
        }

        conn.on('message', function (message) {
          if (message.type === 'utf8') {
            const payload = JSON.parse(message.utf8Data)
            printmsg(self.logger, payload, '  → ', conn.remoteAddress)
            self.emit('recv', payload, conn.remoteAddress)
            self.conductor.recv(payload, self.conn, self.serverConn)
            // self.send(payload, self.conn)
          } else {
            self.logger.warn('wsServer', 'unknown message type', message)
          }
        })

        conn.on('close', function (message) {
          self.logger.info('wsServer', 'connection close', message)
        })
      })

      if (options.catchExceptions) {
        process.on('uncaughtException', function (err) {
          self.send({protocol: 'network', command: 'error', message: err.toString()})
        })
      }
    }
  }

  noCircularJSON (data) {
    return JSON.stringify(data, function (key, val) {
      if (val &&
        typeof val === 'object' &&
        !Array.isArray(val) &&
        val.constructor &&
        val.constructor.name !== 'Object') {
        return '[' + val.constructor.name + ']'
      }
      return val
    })
  }

  /**
   *
   * Send the response.
   *
   */
  send (data, conn) {
    if (data.command === 'error') {
      if (data.payload.message && data.payload.message.message) {
        data.payload.message = data.payload.message.message
      }
    }
    conn.sendUTF(this.noCircularJSON(data))
  }

  static create (http, options) {
    return new WebSocketProxyServerTransport(options, http)
  }
}

module.exports = WebSocketProxyServerTransport
